"""Sayd package."""

# Copyright 2022 LW016 (GPG FINGERPRINT 33CA C1E8 EC4C 0B31 73AE  DD8B 31A5 35D1 2844 39ED)

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ("SaydServer", "SaydClient")
__version__ = "1.2.3"

__author__ = "LW016"
__license__ = "Apache License, Version 2.0"


from ._server import SaydServer
from ._client import SaydClient



"""Server implementation."""

# Copyright 2022 LW016 (GPG FINGERPRINT 33CA C1E8 EC4C 0B31 73AE  DD8B 31A5 35D1 2844 39ED)

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ssl

from time import time
from logging import Logger, getLogger
from typing import Callable, Optional, Union, Dict, Set

from uuid import uuid4
from base64 import b64encode, b64decode
from binascii import Error as BinError

from asyncio import (sleep, create_task, start_server,
        LimitOverrunError, IncompleteReadError, CancelledError)

from asyncio.streams import StreamReader, StreamWriter
from asyncio.events import AbstractServer
from asyncio.tasks import Task


try:
    import ujson as json

except ImportError:
    import json # type: ignore

try:
    from uvloop import install as tune

except ImportError:
    pass


class SaydServer:
    """Server class.

    :param host: A list of addresses or a address to bind the server, defaults to `None`.
    :type host: Optional[Union[str, list]]
    :param port: Port to use, defaults to `7050`.
    :type port: int
    :param queue: Limit of clients waiting to connect, defaults to `512`.
    :type queue: int
    :param limit: Limit of connected clients, defaults to `1024`.
    :type limit: int
    :param buffer: Buffer size limit per client in KiB, defaults to `128`.
    :type buffer: int
    :param timeout: Time in seconds to disconnect a client that is not responding,\
            defaults to `3`.
    :type timeout: int
    :param ping: Frequency in seconds to ping the clients, defaults to `1`.
    :type ping: int
    :param logger: Logger to use, defaults to `None`.
    :type logger: Optional[Logger]
    :param cert: Path to the TLS certificate, defaults to `None`.
    :type cert: Optional[str]
    :param cert_key: Path to the TLS certificate key, defaults to `None`.
    :type cert_key: Optional[str]
    """

    def __init__(
            self,
            host: Optional[Union[str, list]] = None,
            port: int = 7050,
            queue: int = 512,
            limit: int = 1024,
            buffer: int = 128,
            timeout: int = 3,
            ping: int = 1,
            logger: Optional[Logger] = None,
            cert: Optional[str] = None,
            cert_key: Optional[str] = None
            ) -> None:

        self._host = host
        self._port = port

        self._queue_limit = queue
        self._connections_limit = limit
        self._buffer_limit = buffer * 1024
        self._connection_timeout = timeout
        self._ping_timeout = ping


        if logger is not None:
            self._logger = logger

        else:
            self._logger = getLogger()


        self._ssl_context: Union[ssl.SSLContext, bool, None]

        if cert is not None and cert_key is not None:
            self._ssl_context = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            self._ssl_context.load_cert_chain(cert, cert_key)

        else:
            self._ssl_context = None
        

        self._server: AbstractServer
        self._ping_task: Task
        
        self._connections: Dict[tuple, list] = {}
        self._connections_queue: Dict[tuple, list] = {}
        self._blacklist: Set[str] = set()

        self._callbacks: Dict[str, Callable] = {
                "ping": self._ping
                }

        self._call_tasks: Set[Task] = set()
        self._calls: Dict[str, Union[dict, list, None]] = {}

        
        try:
            # Enable uvloop if available.
            tune()

        except NameError:
            pass


    @property
    def clients(self) -> Set[tuple]:
        """Returns the connected clients.

        :return: A set containing the clients.
        :rtype: Set[tuple]
        """

        return set(self._connections.keys())

    
    def callback(self, name: str) -> Callable:
        """Decorator to bind functions to be called when a request is received.

        :param name: Name to bind the function.
        :type name: str
        """

        assert name != "ping", "Name 'ping' is used internally."

        def decorator(function: Callable) -> Callable:
            self._callbacks.update({name: function})

            def wrapper(address: tuple, instance: Union[str, None], data: dict) -> Callable: # pylint: disable=unused-argument
                return function(address, instance, data)

            return wrapper

        return decorator

    
    def add_callback(self, name: str, function: Callable) -> None:
        """Method to bind functions to be called when a request is received.

        :param name: Name to bind the function.
        :type name: str
        :param function: Function to bind.
        :type function: Callable
        """

        assert name != "ping", "Name 'ping' is used internally."

        self._callbacks.update({name: function})


    def blacklist(self, host: str) -> None:
        """Method to add a host to the blacklist.

        :param host: Host to block.
        :type host: str
        """

        self._blacklist.add(host)
        self._logger.info("Server | Host %s added to the blacklist", host)
        
        # Closes connections from the blacklisted host.
        for connection in list(self._connections.keys()):
            if connection[0] == host:
                if not self._connections[connection][0].is_closing():
                    self._connections[connection][0].close()


    def unblacklist(self, host: str) -> None:
        """Method to remove a host from the blacklist.

        :param host: Host to unblock.
        :type host: str
        """

        self._blacklist.discard(host)
        self._logger.info("Server | Host %s removed from the blacklist", host)


    async def call(
            self,
            name: str,
            data: Optional[dict] = None,
            instance: Optional[str] = None,
            address: Optional[tuple] = None,
            wait: bool = True,
            wait_timeout: int = 3,
            _call_id: Optional[str] = None
            ) -> Union[dict, list, bool, Exception, None]:
        """Calls a function in a remote client or in all clients (broadcast) if address
        is not specified.

        :param name: Name of the function.
        :type name: str
        :param data: Data to send, defaults to `None`.
        :type data: dict
        :param instance: Instance to pass to the remote function, defaults to `None`.
        :type instance: Optional[str]
        :param address: Client address, defaults to `None`.
        :type address: Optional[tuple]
        :param wait: `True` to wait for a answer, defaults to `True`.
        :type wait: bool
        :param wait_timeout: Time limit in seconds to wait for a answer, defaults to `3`.
        :type wait_timeout: int
        :return: The call result.
        :rtype: Union[dict, list, bool, Exception, None]
        """
        
        if data is not None:
            assert "call" not in data, "Key 'call' is used internally."
            assert "call_id" not in data, "Key 'call_id' is used internally."
            assert "instance" not in data, "Key 'instance' is used internally."
        
        
        datap: dict = data.copy() if data is not None else {}
        datap.update({
            "call": name,
            "instance": instance
            })

        
        to_wait: bool = False

        if _call_id is not None:
            datap.update({"call_id": _call_id})

        elif wait:
            to_wait = True

            call_id: str = uuid4().hex
            self._calls.update({call_id: [] if address is None else None})

            datap.update({"call_id": call_id})

        
        try:
            dataf: bytes = b64encode(json.dumps(datap).encode("ascii")) + b"&"

            if address is not None:
                if not self._connections[address][0].is_closing():
                    self._connections[address][0].write(dataf)
                    await self._connections[address][0].drain()
                
            else:
                for connection in list(self._connections.keys()):
                    if not self._connections[connection][0].is_closing():
                        self._connections[connection][0].write(dataf)
                        await self._connections[connection][0].drain()

        except (SyntaxError, RuntimeError, KeyError, TypeError,
                ValueError, AssertionError, ConnectionError) as error:

            return error
        
        
        if to_wait:
            start_t: float = time()
            
            # Wait for a answer.
            if address is None:
                con_max: int = len(self._connections)

                while len(self._calls[call_id]) < con_max and (time() - start_t) < wait_timeout: # type: ignore
                    await sleep(0.01)

            else:
                while self._calls[call_id] is None and (time() - start_t) < wait_timeout:
                    await sleep(0.01)

            # If was not answered returns None.
            if self._calls[call_id] is None or (isinstance(self._calls[call_id], list) and len(self._calls[call_id]) == 0): # type: ignore
                del self._calls[call_id]
                return None
            
            # If was answered returns the response.
            answer: Union[dict, list] = self._calls[call_id] # type: ignore
            del self._calls[call_id]
            return answer
        
        return True

    
    async def start(self) -> None:
        """Starts the server."""

        if self._ssl_context is not None:
            ssl_data = {
                    "ssl": self._ssl_context,
                    "ssl_handshake_timeout": self._connection_timeout*2
                    }

        else:
            ssl_data = {}
        
        
        self._server = await start_server(
                self._connection_bootstrap,
                host=self._host, # type: ignore
                port=self._port,
                limit=self._buffer_limit,
                backlog=self._queue_limit,
                start_serving=True,
                **ssl_data) # type: ignore

        self._ping_task = create_task(self._ping_clients())


    async def stop(self) -> None:
        """Stops the server."""

        self._ping_task.cancel()

        # Closes queued connections.
        for connection in list(self._connections_queue.keys()):
            if not self._connections_queue[connection][0].is_closing():
                self._connections_queue[connection][0].close()
            
            self._connections_queue[connection][2].cancel()
            del self._connections_queue[connection]

        # Closes open connections.
        for connection in list(self._connections.keys()):
            if not self._connections[connection][0].is_closing():
                self._connections[connection][0].close()
            
            self._connections[connection][2].cancel()
            del self._connections[connection]

        self._server.close()

    
    async def _ping_clients(self) -> None:
        """Continuously check the connection status of the clients."""

        while 1:
            current_time = time()

            # Checks queued connections.
            for connection in list(self._connections_queue.keys()):
                # Checks if the connection is closed.
                if self._connections_queue[connection][0].is_closing():
                    self._connections_queue[connection][2].cancel()
                    del self._connections_queue[connection]

                # Checks if the client stopped responding.
                elif (current_time - self._connections_queue[connection][1]) >= self._connection_timeout: 
                    if not self._connections_queue[connection][0].is_closing():
                        self._connections_queue[connection][0].close()

                    self._connections_queue[connection][2].cancel()
                    del self._connections_queue[connection]

            # Checks open connections.
            for connection in list(self._connections.keys()):
                # Checks if the connection is closed.
                if self._connections[connection][0].is_closing():
                    self._connections[connection][2].cancel()
                    del self._connections[connection]

                    self._logger.info("Server | Disconnection from %s:%s", connection[0], connection[1])
                
                # Checks if the client stopped responding.
                elif (current_time - self._connections[connection][1]) >= self._connection_timeout: 
                    if not self._connections[connection][0].is_closing():
                        self._connections[connection][0].close()

                    self._connections[connection][2].cancel()
                    del self._connections[connection]

                    self._logger.info("Server | Disconnection from %s:%s", connection[0], connection[1])
                
                # Ping the client.
                else:
                    await self.call(name="ping", address=connection, wait=False)

            await sleep(self._ping_timeout)

    
    async def _ping(self, address: tuple, instance: None, data: dict) -> None: # pylint: disable=unused-argument
        """Called when a ping is received from a client."""

        try:
            self._connections[address][1] = time()

        except KeyError as error:
            self._logger.error("Server | Ping receive error (%s)", error)


    async def _connection_bootstrap(self, reader: StreamReader, writer: StreamWriter) -> None:
        address: tuple = writer.get_extra_info("peername")

        
        if address is None or address[0] in self._blacklist or address in self._connections_queue\
                or address in self._connections:

            writer.close()

        elif (len(self._connections)+len(self._connections_queue))+1 > self._connections_limit:
            writer.close()

            self._logger.info("Server | Connection rejected from %s:%s (max capacity)", address[0], address[1])

        else:
            task: Task = create_task(self._connection_handler(reader, address))

            self._connections_queue.update({address: [writer, time(), task]})
 

    async def _connection_handler(self, reader: StreamReader, address: tuple) -> None:
        stream_reader = reader

        recv_data: Union[bytes, str]
        raw_data: str
        data: dict
        
        call: str
        call_id: Union[str, None]
        instance: str
        task: Task

        queued: bool = True
        header: Union[str, list]
        header_addr: tuple


        async def reset_buffer(timeout: float = 0) -> None:
            try:
                del recv_data, raw_data, data, call, call_id, instance, task

            except NameError:
                pass

            await sleep(timeout)


        while 1:
            try:
                try:
                    recv_data = await stream_reader.readuntil(b"&")

                except LimitOverrunError:
                    self._logger.info("Server | Buffer limit exceeded for host %s:%s", address[0], address[1])

                    await reset_buffer(0.1)
                    continue

                except (IncompleteReadError, ConnectionError):
                    await reset_buffer(0.1)
                    continue


                if queued:
                    raw_data = recv_data.decode("ascii")

                    # Checks if the data has a proxy protocol header.
                    if raw_data.find("PROXY") != -1:
                        header, recv_data = raw_data.split("\n")
                        recv_data = recv_data.encode("ascii")

                        header = header.split(" ")
                        header_addr = (header[2], int(header[4]))
                        
                        # Verify if the address is valid.
                        if header_addr[0] in self._blacklist or header_addr in self._connections:
                            self._connections_queue[address][0].close()
                            return
                        
                        self._connections.update({header_addr: self._connections_queue[address]})
                        del self._connections_queue[address]

                        address = header_addr
                    
                    elif raw_data.find("UNKNOWN") != -1:
                        self._connections_queue[address][0].close()
                        return

                    else:
                        self._connections.update({address: self._connections_queue[address]})
                        del self._connections_queue[address]
                    
                    self._logger.info("Server | Connection from %s:%s", address[0], address[1])
                    queued = False


                raw_data = b64decode(recv_data[:-1].decode("ascii")) # type: ignore
                data = json.loads(raw_data)


                if "call_id" in data:
                    call_id = data.pop("call_id")
                    
                    # Checks if the data is a answer of a previous call.
                    if call_id in self._calls:
                        del data["call"]
                        del data["instance"]
                        
                        data.update({"address": address})

                        if isinstance(self._calls[call_id], list):
                            self._calls[call_id].append(data) # type: ignore

                        else:
                            self._calls[call_id] = data

                        await reset_buffer()
                        continue

                else:
                    call_id = None


                call = data.pop("call")
                instance = data.pop("instance")


                if call in self._callbacks:
                    task = create_task(self._callbacks[call](address, instance, data))

                    setattr(task, "call_id", call_id)
                    setattr(task, "call_name", call)
                    setattr(task, "call_instance", instance)
                    setattr(task, "call_address", address)

                    self._call_tasks.add(task)
                    task.add_done_callback(self._on_call_exit)

                elif "*" in self._callbacks:
                    task = create_task(self._callbacks["*"](address, instance, data))
                    
                    setattr(task, "call_id", call_id)
                    setattr(task, "call_name", call)
                    setattr(task, "call_instance", instance)
                    setattr(task, "call_address", address)

                    self._call_tasks.add(task)
                    task.add_done_callback(self._on_call_exit)

            
            except (json.JSONDecodeError, BinError, KeyError, TypeError,
                    ValueError, IndexError, UnicodeDecodeError) as error:
                
                self._logger.error("Server | Error in the data received (%s)", error)

                await reset_buffer(0.1)
                continue

            await reset_buffer()


    def _on_call_exit(self, task: Task) -> None:
        result: Union[None, dict]

        try:
            result = task.result()

        except CancelledError:
            result = None

        if result is not None and isinstance(result, dict) and task.call_id is not None: # type: ignore
            try:
                rtask: Task = create_task(self.call(
                    name=task.call_name, # type: ignore
                    data=result,
                    instance=task.call_instance, # type: ignore
                    address=task.call_address, # type: ignore
                    wait=False,
                    _call_id=task.call_id # type: ignore
                    ))

                self._call_tasks.add(rtask)
                rtask.add_done_callback(self._call_tasks.discard)

            except AssertionError:
                pass

        self._call_tasks.discard(task)



"""CLI utility to generate self-signed certificates."""

# Copyright 2022 LW016 (GPG FINGERPRINT 33CA C1E8 EC4C 0B31 73AE  DD8B 31A5 35D1 2844 39ED)

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional
from OpenSSL import crypto
from typer import run # type: ignore


def main(
        cert_output: str = "cert.crt",
        key_output: str = "cert.key",
        country: str = "BR",
        state: str = "Sao Paulo",
        locality: str = "Sao Paulo",
        org: str = "Local",
        org_unity: str = "Local",
        common_name: str = "localhost",
        exp_days: int = 365,
        email: Optional[str] = None
        ) -> None:
    """Generates a self-signed certificate."""

    # Create a key pair
    key = crypto.PKey()
    key.generate_key(crypto.TYPE_RSA, 2048)

    
    # Create a self-signed cert
    cert = crypto.X509()
    cert.get_subject().C = country
    cert.get_subject().ST = state
    cert.get_subject().L = locality
    cert.get_subject().O = org
    cert.get_subject().OU = org_unity
    cert.get_subject().CN = common_name

    if email is not None:
        cert.get_subject().emailAddress = email

    
    cert.set_serial_number(1000)

    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(exp_days * ((60*60)*24))

    
    cert.set_issuer(cert.get_subject())
    cert.set_pubkey(key)
    cert.sign(key, 'sha256')

    
    # Save certificate
    with open(cert_output, "wb") as file:
        file.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))

    
    # Save key
    with open(key_output, "wb") as file:
        file.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, key))


def execute() -> None:
    """Script entry point."""

    run(main)


if __name__ == "__main__":
    execute()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sayd']

package_data = \
{'': ['*']}

install_requires = \
['pyOpenSSL>=22.0.0', 'typer>=0.4.1']

entry_points = \
{'console_scripts': ['sayd = sayd.__main__:execute']}

setup_kwargs = {
    'name': 'sayd',
    'version': '1.2.3',
    'description': 'A performant asynchronous communication protocol in pure Python.',
    'long_description': 'Sayd\n====\n*A performant asynchronous communication protocol in pure Python.*\n\nThis library was developed with simplicity and performance in mind, with modern practices of Python development, currently in a test state.\n\n`Documentation Reference <https://sayd.readthedocs.io>`_\n\n\nInstall\n-------\nWorks on Python 3.7.4+.\n\n.. code-block:: bash\n\n    pip install sayd\n\n\nDevelopment\n-----------\nYou need to have installed `poetry <https://github.com/python-poetry/poetry>`_ for dependencies management (`how to <https://python-poetry.org/docs/#installation>`_).\n\n.. code-block:: bash\n\n    git clone https://github.com/lw016/sayd\n    cd sayd\n    poetry install\n\n\nRun tests\n^^^^^^^^^\n.. code-block:: bash\n\n    poetry run tox -e tests\n\nBuild docs\n^^^^^^^^^^\n.. code-block:: bash\n\n    poetry run tox -e docs\n\n\nFeatures\n--------\n- Client and server implementations\n- Reliable TCP persistent connection\n- Auto reconnection *(client)*\n- Multiple asynchronous connections *(server)*\n- Blacklist of clients *(server)*\n- TLS encryption\n- Proxy Protocol V2 support *(server)*\n- Data transmitted as dictionaries *(json)*\n- Broadcast *(server)*\n- Remote function callbacks\n- Built-in CLI utility to generate self-signed certificates\n\n\nRoadmap\n-------\n- Add support to Unix socket\n- Implement TLS certificate authentication\n\n\nCLI\n---\nThe built-in CLI utility (*sayd*) can be used to generate self-signed certificates to encrypt the connection.\n\n.. code-block:: bash\n\n    sayd --help\n\n\nUsage example\n-------------\nServer\n^^^^^^\n.. code-block:: python\n\n    import logging\n    import asyncio\n\n    from sayd import SaydServer\n\n\n    logging.basicConfig(\n            format="[%(name)s][%(levelname)s] %(asctime)s - %(message)s",\n            datefmt="%Y/%m/%d %H:%M:%S"\n            )\n\n    logger = logging.getLogger("SERVER")\n    logger.setLevel(logging.INFO)\n\n\n    server = SaydServer(logger=logger)\n\n\n    @server.callback("message")\n    async def msg(address: tuple, instance: str, data: dict) -> dict:\n        return {"greetings": "Hello from server!"}\n\n\n    async def main() -> None:\n        await server.start()\n        \n        \n        while True:\n            result = await server.call("msg", {"greetings": "Hi!"}) # Broadcast call.\n            print(result)\n\n            await asyncio.sleep(1)\n        \n        \n        await server.stop()\n\n\n    if __name__ == "__main__":\n        asyncio.run(main())\n\nClient\n^^^^^^\n.. code-block:: python\n\n    import logging\n    import asyncio\n\n    from sayd import SaydClient\n\n\n    logging.basicConfig(\n            format="[%(name)s][%(levelname)s] %(asctime)s - %(message)s",\n            datefmt="%Y/%m/%d %H:%M:%S"\n            )\n\n    logger = logging.getLogger("CLIENT")\n    logger.setLevel(logging.INFO)\n\n\n    client = SaydClient(logger=logger)\n\n\n    @client.callback("msg")\n    async def msg(instance: str, data: dict) -> dict:\n        return {"greetings": "Hello from client!"}\n\n\n    async def main() -> None:\n        await client.start()\n\n\n        while True:\n            result = await client.call("message", {"greetings": "Hi!"})\n            print(result)\n\n            await asyncio.sleep(1)\n\n        \n        await client.stop()\n\n\n    if __name__ == "__main__":\n        asyncio.run(main())\n',
    'author': 'LW016',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/lw016/sayd',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.4,<4.0.0',
}


setup(**setup_kwargs)



"""Client implementation."""

# Copyright 2022 LW016 (GPG FINGERPRINT 33CA C1E8 EC4C 0B31 73AE  DD8B 31A5 35D1 2844 39ED)

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

# http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ssl

from time import time
from logging import Logger, getLogger
from typing import Callable, Optional, Union, Dict, Set

from uuid import uuid4
from base64 import b64encode, b64decode
from binascii import Error as BinError

from asyncio import (sleep, create_task, open_connection,
        LimitOverrunError, IncompleteReadError, CancelledError)

from asyncio.streams import StreamReader, StreamWriter
from asyncio.tasks import Task


try:
    import ujson as json

except ImportError:
    import json # type: ignore

try:
    from uvloop import install as tune

except ImportError:
    pass


class SaydClient:
    """Client class.
    
    :param host: Server hostname, defaults to localhost.
    :type host: str
    :param port: Server port, defaults to `7050`.
    :type port: int
    :param local_host: Local address to bind the client, defaults to `None`.
    :type local_host: str
    :param local_port: Local port to bind the client, defaults to `None`.
    :type local_port: Optional[Union[int, str]]
    :param buffer: Buffer size limit in KiB, defaults to `128`.
    :type buffer: int
    :param timeout: Time in seconds to disconnect from the server if is not responding,\
            defaults to `3`.
    :type timeout: int
    :param ping: Frequency in seconds to ping the server, defaults to `1`.
    :type ping: int
    :param reconnect: If disconnected from the server try to reconnect, defaults to `True`.
    :type reconnect: bool
    :param reconnect_timeout: Frequency in seconds to try to reconnect if disconnected,\
            defaults to `5`.
    :type reconnect_timeout: int
    :param logger: Logger to use, defaults to `None`.
    :type logger: Optional[Logger]
    :param cert: Path to the TLS certificate, defaults to `None`.
    :type cert: Optional[str]
    """

    def __init__(
            self,
            host: str = "localhost",
            port: int = 7050,
            local_host: Optional[str] = None,
            local_port: Optional[Union[int, str]] = None,
            buffer: int = 128,
            timeout: int = 3,
            ping: int = 1,
            reconnect: bool = True,
            reconnect_timeout: int = 5,
            logger: Optional[Logger] = None,
            cert: Optional[str] = None
            ) -> None:

        self._host = host
        self._port = port
        self._dlocal_host = local_host
        self._dlocal_port = local_port
        
        self._buffer_limit = buffer * 1024
        self._connection_timeout = timeout
        self._ping_timeout = ping

        self._reconnect = reconnect
        self._reconnect_timeout = reconnect_timeout


        if logger is not None:
            self._logger = logger

        else:
            self._logger = getLogger()

        
        self._ssl_context: Union[ssl.SSLContext, bool, None]

        if cert is not None:
            self._ssl_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH, cafile=cert)

        else:
            self._ssl_context = None

        
        self._writer: StreamWriter
        self._reader: StreamReader
        self._client_task: Task
        self._ping_task: Task
        
        self._reconnect_state: bool = False
        self._connection_time: float

        self._callbacks: Dict[str, Callable] = {
                "ping": self._ping
                }

        self._call_tasks: Set[Task] = set()
        self._calls: Dict[str, Union[dict, None]] = {}


        try:
            # Enable uvloop if available.
            tune()

        except NameError:
            pass


    @property
    def connected(self) -> bool:
        """Returns the connection status.

        :return: `True` if connected, else `False`.
        :rtype: bool
        """

        return not self._writer.is_closing()

    
    def callback(self, name: str) -> Callable:
        """Decorator to bind functions to be called when a request is received.

        :param name: Name to bind the function.
        :type name: str
        """

        assert name != "ping", "Name 'ping' is used internally."

        def decorator(function: Callable) -> Callable:
            self._callbacks.update({name: function})
            
            def wrapper(instance: Union[str, None], data: dict) -> Callable: # pylint: disable=unused-argument
                return function(instance, data)

            return wrapper

        return decorator

    
    def add_callback(self, name: str, function: Callable) -> None:
        """Method to bind functions to be called when a request is received.

        :param name: Name to bind the function.
        :type name: str
        :param function: Function to bind.
        :type function: Callable
        """

        assert name != "ping", "Name 'ping' is used internally."

        self._callbacks.update({name: function})


    async def call(
            self,
            name: str,
            data: Optional[dict] = None,
            instance: Optional[str] = None,
            wait: bool = True,
            wait_timeout: int = 3,
            _call_id: Optional[str] = None
            ) -> Union[dict, bool, Exception, None]:
        """Calls a function in the server.

        :param name: Name of the function.
        :type name: str
        :param data: Data to send, defaults to `None`.
        :type data: dict
        :param instance: Instance to pass to the remote function, defaults to `None`.
        :type instance: Optional[str]
        :param wait: `True` to wait for a answer, defaults to `True`.
        :type wait: bool
        :param wait_timeout: Time limit in seconds to wait for a answer, defaults to `3`.
        :type wait_timeout: int
        :return: The call result.
        :rtype: Union[dict, bool, Exception, None]
        """
        
        if data is not None:
            assert "call" not in data, "Key 'call' is used internally."
            assert "call_id" not in data, "Key 'call_id' is used internally."
            assert "instance" not in data, "Key 'instance' is used internally."


        datap: dict = data.copy() if data is not None else {}
        datap.update({
            "call": name,
            "instance": instance
            })


        to_wait: bool = False

        if _call_id is not None:
            datap.update({"call_id": _call_id})

        elif wait:
            to_wait = True

            call_id: str = uuid4().hex
            self._calls.update({call_id: None})

            datap.update({"call_id": call_id})
 
        
        try:
            dataf: bytes = b64encode(json.dumps(datap).encode("ascii")) + b"&"

            if not self._writer.is_closing():
                self._writer.write(dataf)
                await self._writer.drain()

        except (SyntaxError, RuntimeError, TypeError,
                ValueError, AssertionError, ConnectionError) as error:

            return error

        
        if to_wait:
            start_t: float = time()
            
            # Wait for a answer.
            while self._calls[call_id] is None and (time() - start_t) < wait_timeout:
                await sleep(0.01)

            # If was not answered returns None.
            if self._calls[call_id] is None:
                del self._calls[call_id]
                return None
            
            # If was answered returns the response.
            answer: dict = self._calls[call_id] # type: ignore
            del self._calls[call_id]
            return answer
        
        return True


    async def start(self) -> None:
        """Starts the client."""

        if self._ssl_context is not None:
            ssl_data = {
                    "ssl": self._ssl_context,
                    "ssl_handshake_timeout": self._connection_timeout*2
                    }

        else:
            ssl_data = {}


        if self._dlocal_host is None or self._dlocal_port is None:
            self._reader, self._writer = await open_connection( # type: ignore
                    host=self._host,
                    port=self._port,
                    limit=self._buffer_limit,
                    **ssl_data) # type: ignore

        # Binds to a specific local host and port.
        else:
            self._reader, self._writer = await open_connection( # type: ignore
                    host=self._host,
                    port=self._port,
                    local_addr=(self._dlocal_host, self._dlocal_port),
                    limit=self._buffer_limit,
                    **ssl_data) # type: ignore

        self._logger.info("Client | Connected to %s:%s", self._host, self._port)


        self._connection_time = time()

        if not self._reconnect_state:
            self._ping_task = create_task(self._ping_server())
            self._client_task = create_task(self._connection_handler())
            self._reconnect_state = True

        
    async def stop(self) -> None:
        """Stops the client."""

        self._ping_task.cancel()
        self._client_task.cancel()
        self._writer.close()

        self._reconnect_state = False
        self._connection_time = 0

    
    async def _ping_server(self) -> None:
        """Continuously check the connection status of the server."""

        while 1:
            current_time = time()

            # Checks if the connection is closed or if the server stop responding.
            if self._writer.is_closing() or ((current_time - self._connection_time) >= self._connection_timeout):
                if not self._writer.is_closing():
                    self._writer.close()

                self._connection_time = 0

                self._logger.info("Client | Disconnected from %s:%s", self._host, self._port)

                # Will keep trying to reconnect in case of disconnection.
                while self._reconnect and not self._connection_time:
                    self._logger.info("Client | Trying to reconnect to %s:%s", self._host, self._port)
                    
                    try:
                        await self.start()

                    except ConnectionError:
                        await sleep(self._reconnect_timeout)

                    else:
                        break
            
            # Ping the server.
            else:
                await self.call(name="ping", wait=False)

            await sleep(self._ping_timeout)
    
    
    async def _ping(self, instance: None, data: dict) -> None: # pylint: disable=unused-argument
        """Called when a ping is received from the server."""

        self._connection_time = time()
            

    async def _connection_handler(self) -> None:
        recv_data: bytes
        raw_data: str
        data: dict
        
        call: str
        call_id: Union[str, None]
        instance: str
        task: Task

        
        async def reset_buffer(timeout: float = 0) -> None:
            try:
                del recv_data, raw_data, data, call, call_id, instance, task

            except NameError:
                pass

            await sleep(timeout)


        while 1:
            try:
                try:
                    recv_data = await self._reader.readuntil(b"&")

                except LimitOverrunError:
                    self._logger.info("Client | Buffer limit exceeded")

                    await reset_buffer(0.1)
                    continue

                except (IncompleteReadError, ConnectionError):
                    await reset_buffer(0.1)
                    continue


                raw_data = b64decode(recv_data[:-1].decode("ascii")) # type: ignore
                data = json.loads(raw_data)


                if "call_id" in data:
                    call_id = data.pop("call_id")
                    
                    # Checks if the data is a answer of a previous call.
                    if call_id in self._calls:
                        del data["call"]
                        del data["instance"]

                        self._calls[call_id] = data

                        await reset_buffer()
                        continue

                else:
                    call_id = None


                call = data.pop("call")
                instance = data.pop("instance")


                if call in self._callbacks:
                    task = create_task(self._callbacks[call](instance, data))
                    
                    setattr(task, "call_id", call_id)
                    setattr(task, "call_name", call)
                    setattr(task, "call_instance", instance)

                    self._call_tasks.add(task)
                    task.add_done_callback(self._on_call_exit)

                elif "*" in self._callbacks:
                    task = create_task(self._callbacks["*"](instance, data))
                    
                    setattr(task, "call_id", call_id)
                    setattr(task, "call_name", call)
                    setattr(task, "call_instance", instance)

                    self._call_tasks.add(task)
                    task.add_done_callback(self._on_call_exit)

                
            except (json.JSONDecodeError, BinError, KeyError, TypeError,
                    ValueError, UnicodeDecodeError) as error:

                self._logger.error("Client | Error in the data received (%s)", error)

                await reset_buffer(0.1)
                continue

            await reset_buffer()


    def _on_call_exit(self, task: Task) -> None:
        result: Union[None, dict]

        try:
            result = task.result()

        except CancelledError:
            result = None

        if result is not None and isinstance(result, dict) and task.call_id is not None: # type: ignore
            try:
                rtask: Task = create_task(self.call(
                    name=task.call_name, # type: ignore
                    data=result,
                    instance=task.call_instance, # type: ignore
                    wait=False,
                    _call_id=task.call_id # type: ignore
                    ))

                self._call_tasks.add(rtask)
                rtask.add_done_callback(self._call_tasks.discard)

            except AssertionError:
                pass

        self._call_tasks.discard(task)
